import pytest

from eunomia import eunomia_load
from eunomia.config import Group, Option
from eunomia.registry import RegistryGroup


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def foo(bar, baz=1):
    return bar, baz


def test_simple_option():
    root = RegistryGroup()

    _foo = foo
    _foo = root.register()(_foo)
    _foo = root.register(option_name='foo_alt')(_foo)
    _foo = root.register(option_name='foo_alt2', group_path='/group/subgroup')(_foo)
    _foo = root.register(option_name='foo_alt3', group_path='/group/subgroup', override_defaults=dict(baz=2))(_foo)
    assert _foo is foo

    target = Group({
        'tests': Group({
            'test_registry': Group({
                'foo': Option({'_target_': 'tests.test_registry.foo', 'baz': 1}),
                'foo_alt': Option({'_target_': 'tests.test_registry.foo', 'baz': 1}),
            })
        }),
        'group': Group({
            'subgroup': Group({
                'foo_alt2': Option({'_target_': 'tests.test_registry.foo', 'baz': 1}),
                'foo_alt3': Option({'_target_': 'tests.test_registry.foo', 'baz': 2}),
            })
        })
    })

    # NOTE: the working directory must not be /eunomia/tests, it must be /eunomia
    assert root.to_dict() == target.to_dict()

    # test defaults
    # - if is_default is not specified, then the default can be overwritten
    #   otherwise it cannot be

    assert root.get_registered_defaults() == {'tests/test_registry': 'foo'}
    assert root.get_registered_defaults(explicit_only=True) == {}

    _foo = root.register(option_name='foo_alt4', group_path='/group/subgroup', override_defaults=dict(baz=3), is_default=True)(_foo)

    assert root.get_registered_defaults() == {'group/subgroup': 'foo_alt4'}
    assert root.get_registered_defaults(explicit_only=True) == {'group/subgroup': 'foo_alt4'}

    # test that the registrable group is the root.
    g = Group({'temp': root})
    with pytest.raises(AssertionError, match='Can only register on the root node.'): root.register()(_foo)
    with pytest.raises(AssertionError, match='Can only register on the root node.'): root.get_registered_defaults()

    # reset
    root._parent = None
    root._key = None

    # try again
    # TODO: maybe update to better error message
    with pytest.raises(KeyError, match='parent already has child with key: foo'): root.register()(_foo)
    assert root.get_registered_defaults() == {'group/subgroup': 'foo_alt4'}

    # try merge group
    root.add_option('default', Option(defaults=root.get_registered_defaults()))
    assert eunomia_load(root) == {'group': {'subgroup': {'_target_': 'tests.test_registry.foo', 'baz': 3}}}


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from eunomia.backend import Backend
from eunomia.config import Group


# ========================================================================= #
# Internal Backend                                                          #
# ========================================================================= #


class BackendObj(Backend):
    """
    The most basic backend is the ConfigObject data structures
    used by the core config loader.

    The backend passes through the values as they
    are without any modifications.
    """

    def __init__(self, root_group: Group):
        if not isinstance(root_group, Group):
            raise TypeError(f'root_group must be an instance of {Group.__name__}')
        self._root_group = root_group

    def _load_root_group(self) -> Group:
        return self._root_group




import inspect
import os
import re
from collections import defaultdict
from typing import Dict, Union, DefaultDict, List, Tuple, Optional

from eunomia.config import Group, Option
from eunomia.config import scheme as s


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

def _get_module_path(obj):
    path = os.path.relpath(inspect.getmodule(obj).__file__, os.getcwd())
    assert path.endswith('.py')
    path = path[:-len('.py')]
    return os.path.normpath(path)


def _get_import_path(obj):
    module_path = '.'.join(_get_module_path(obj).split('/'))
    return f'{module_path}.{obj.__name__}'


def _camel_to_snake(name):
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


# ========================================================================= #
# Advanced Group                                                            #
# ========================================================================= #


class RegistryGroup(Group):

    """
    This group allows registration of functions or classes, constructing an
    option that has the _target_ parameter set, with default values obtained
    from the default parameters of the function.

    The option by default resides under the group that corresponds to the module
    that the function or class resides in.
    """

    def __init__(
            self,
            named_nodes: Dict[str, Union['Group', 'Option']] = None,
    ):
        super().__init__(named_nodes=named_nodes)
        # save all default options
        self._registered_all: DefaultDict[object, List[Option]] = defaultdict(list)
        self._registered_defaults: Dict[object, Tuple[bool, Option]] = {}

    def register(self, option_name: str = None, group_path: str = None, target: str = None, override_defaults=None, is_default=None):
        assert not self.has_parent, 'Can only register on the root node.'
        # wrap!
        def wrapper(func):
            # get the function paths
            import_path = _get_import_path(func) if target is None else target
            module_path = _get_module_path(func) if group_path is None else group_path
            # construct option
            overrides = override_defaults if override_defaults else {}
            kwargs = _get_default_args(func)
            for k in kwargs:
                if k in overrides:
                    kwargs[k] = overrides.pop(k)
            assert not overrides, f'tried to override missing values: {list(overrides.keys())}'
            assert s.MARKER_KEY_TARGET not in kwargs, f'object {import_path} has conflicting optional parameter: {s.MARKER_KEY_TARGET}'
            # get group and make missing along path
            group = self.get_group_from_path(path=module_path, make_missing=True)
            # get the name of the option from the function
            option_key = _camel_to_snake(func.__name__) if option_name is None else option_name
            # register the option
            option = group.add_option(option_key, Option(data={
                s.MARKER_KEY_TARGET: import_path,
                **kwargs
            }))
            self._register_obj(func, option, is_default)
            # return the original function
            return func

        # decorate correctly!
        if callable(group_path):
            fn, group_path = group_path, None
            return wrapper(fn)
        else:
            return wrapper

    def _register_obj(self, func, option, is_default=None):
        assert not self.has_parent, 'Can only register on the root node.'
        # keep track that this was registered
        self._registered_all[func].append(option)
        # register as a default
        if func in self._registered_defaults:
            explicit, _ = self._registered_defaults[func]
            if explicit:
                if is_default:
                    raise AssertionError(f'registered callable {func} for option: {option.keys} was previously explicitly registered as a default.')
            else:
                if is_default:
                    self._registered_defaults[func] = (True, option)
                elif is_default is None:
                    self._registered_defaults.setdefault(func, (False, option))
        else:
            if is_default or (is_default is None):
                self._registered_defaults.setdefault(func, (False if is_default is None else True, option))

    def get_registered_defaults(self, explicit_only=False):
        assert not self.has_parent, 'Can only register on the root node.'
        # get default options!
        defaults = {}
        for func, (explicit, option) in self._registered_defaults.items():
            if explicit_only and not explicit:
                continue
            k = option.group_path
            if k in defaults:
                raise AssertionError(f'default for callable {func} corresponding to option: {option.keys} has already been added.')
            defaults[k] = option.key
        return defaults

# ========================================================================= #
# END                                                                       #
# ========================================================================= #




from pathlib import Path
from typing import Union

from eunomia.backend import BackendYaml, BackendDict, BackendObj
from eunomia.config import Group


DefaultConfigTypes = Union[str, Path, dict, Group]


def make_default_backend(config: DefaultConfigTypes):
    if isinstance(config, (str, Path)):
        # we assume the config is a path to the root folder for a YAML backend
        if isinstance(config, Path):
            config = str(config.absolute())
        return BackendYaml(root_folder=config)
    elif isinstance(config, dict):
        # we assume that the config is a dictionary
        return BackendDict(root_dict=config)
    elif isinstance(config, Group):
        # we assume that the config is a ConfigGroup with ConfigOptions
        return BackendObj(root_group=config)
    else:
        raise TypeError(f'Unsupported config_root type: {config.__class__.__name__}')



from typing import Dict, Union, List, Tuple

from eunomia._util_traverse import RecursiveTransformer
from eunomia.config import scheme as s
from eunomia.config.nodes import ConfigNode, SubNode


# ========================================================================= #
# Config Node                                                               #
# ========================================================================= #


class _ConfigObject(object):

    def __init__(self):
        super().__init__()
        self._parent = None
        self._key = None
        self._children = {}

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Path                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def has_parent(self):
        assert not ((not self._key) ^ (not self._parent))
        return (self._key is not None) and (self._parent is not None)

    @property
    def parent(self) -> '_ConfigObject':
        assert self.has_parent
        return self._parent

    @property
    def key(self) -> str:
        assert self.has_parent
        return self._key

    @property
    def keys(self) -> Tuple[str]:
        return tuple(n.key for n in self.walk_from_root(visit_root=False))

    @property
    def root(self) -> '_ConfigObject':
        node = self
        for node in self.walk_to_root(visit_root=True):
            pass
        return node

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Walk                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def walk_to_root(self, visit_root=False):
        current = self
        while current.has_parent:
            yield current
            current = current.parent
        if visit_root:
            yield current

    def walk_from_root(self, visit_root=False):
        yield from reversed(list(self.walk_to_root(
            visit_root=visit_root
        )))

    def walk_descendants(self):
        def _iter(node):
            yield node
            for k in node:
                yield from _iter(node[k])
        return _iter(self)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Children                                                              #
    # - for simplicity we do not allow attribute access to children.        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def add_child(self, key: str, child: '_ConfigObject') -> '_ConfigObject':
        if not isinstance(child, _ConfigObject):
            raise TypeError(f'child must be an instance of {_ConfigObject.__name__}')
        if child.has_parent:
            raise ValueError(f'child already has a parent, and cannot be added.')
        if key in self._children:
            raise KeyError(f'parent already has child with key: {key}')
        # set details
        child._parent = self
        child._key = key
        self._children[key] = child
        # return the added value
        return child

    def get_child(self, key: str):
        return self._children[key]

    def has_child(self, key: str):
        return key in self._children

    __setitem__ = add_child
    __getitem__ = get_child
    __contains__ = has_child

    def __iter__(self):
        yield from self._children

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Schema                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def from_dict(cls, mapping: dict, validate=True):
        raise NotImplementedError

    def to_dict(self, validate=True):
        raise NotImplementedError

    def is_valid_dict(self):
        try:
            self.to_dict()
            return True
        except:
            return False

    def to_yaml(self) -> str:
        import ruamel.yaml
        return ruamel.yaml.round_trip_dump(self.to_dict())


# ========================================================================= #
# Group                                                                     #
# ========================================================================= #


class Group(_ConfigObject):

    def __init__(
            self,
            named_nodes: Dict[str, Union['Group', 'Option']] = None,
    ):
        super().__init__()
        # add groups or options
        if named_nodes:
            for k, v in named_nodes.items():
                self.add_child(k, v)
            # we dont need to validate groups because we dont
            # directly support conversion to them yet.

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Override                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def add_child(self, key, node: Union['Group', 'Option']) -> Union['Group', 'Option']:
        if not isinstance(node, (Group, Option)):
            raise TypeError(f'Can only add {Group.__name__} and {Option.__name__} nodes to {Group.__name__}')
        # add!
        assert super().add_child(key, node) is node
        # we dont return the above because of type warnings
        return node

    # override to yield in order
    def __iter__(self):
        yield from self.groups
        yield from self.options

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Getters                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def root(self) -> 'Group':
        g = super().root
        assert isinstance(g, Group)
        return g

    @property
    def groups(self) -> Dict[str, 'Group']:
        return {k: v for k, v in self._children.items() if isinstance(v, Group)}

    @property
    def options(self) -> Dict[str, 'Option']:
        return {k: v for k, v in self._children.items() if isinstance(v, Option)}

    @property
    def group_path(self):
        return '/'.join(self.keys if self.keys else ())

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Groups & Options                                                      #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_subgroup(self, key: str) -> 'Group':
        node = self.get_child(key)
        if not isinstance(node, Group):
            raise TypeError(f'node with key: {repr(key)} is not a {Group.__name__}')
        return node

    def get_option(self, key: str) -> 'Option':
        node = self.get_child(key)
        if not isinstance(node, Option):
            raise TypeError(f'node with key: {repr(key)} is not a {Option.__name__}')
        return node

    def add_subgroup(self, key: str, value: 'Group') -> 'Group':
        if not isinstance(value, Group):
            raise TypeError(f'adding node with key: {repr(key)} is not a {Group.__name__}')
        return self.add_child(key, value)

    def add_option(self, key: str, value: 'Option') -> 'Option':
        if not isinstance(value, Option):
            raise TypeError(f'adding node with key: {repr(key)} is not a {Option.__name__}')
        return self.add_child(key, value)

    def new_subgroup(self, key: str) -> 'Group':
        return self.add_subgroup(key, Group())

    # we don't support new_option because we shouldn't
    # modify it after it is initialised
    # def new_option(self, key: str) -> 'Option':
    #     return self.add_option(key, Option())

    def has_suboption(self, key):
        if not self.has_child(key):
            return False
        if not isinstance(self.get_child(key), Option):
            return False
        return True

    def has_subgroup(self, key):
        if not self.has_child(key):
            return False
        if not isinstance(self.get_child(key), Group):
            return False
        return True

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Walk                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def get_subgroups_recursive(self, keys: List[str], make_missing=False) -> 'Group':
        def _recurse(group: Group, old_keys):
            if not old_keys:
                return group
            key, keys = old_keys[0], old_keys[1:]
            if make_missing:
                if not group.has_subgroup(key):
                    return _recurse(group.new_subgroup(key), keys)
            return _recurse(group.get_subgroup(key), keys)
        return _recurse(self, keys)

    def get_group_from_path(self, path: str, make_missing=False):
        """
        This function checks if a path is relative or absolute.
        If the path is relative, it traverses from the current group.
        If the path is absolute, it traverses from the root group.

        Supports make_missing like get_subgroups_recursive(...)
        """
        group_keys, is_relative = s.split_group_path(path)
        # get the group corresponding to the path - must handle relative & root paths
        root = (self if is_relative else self.root)
        return root.get_subgroups_recursive(group_keys, make_missing=make_missing)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Strings                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self._children})'

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Debug                                                                 #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def debug_print_tree(self, colors=True, show_options=True, full_option_path=True, full_group_path=True):
        from attr import dataclass

        @dataclass
        class _WalkObj:
            node: _ConfigObject
            visited: bool
            is_last: bool
            @property
            def is_group(self):
                return isinstance(self.node, Group)
            @property
            def has_groups_after(self):
                return self.node.has_parent and isinstance(self.node.parent, Group) and bool(self.node.parent.groups)
            @property
            def key(self):
                return (self.is_group, self.visited, self.is_last, self.has_groups_after)
            def __str__(self):
                return f":{''.join(map(str, map(int, self.key)))}"

        def _is_last_iter(items):
            yield from ((item, i == len(items)-1) for i, item in enumerate(items))

        def _walk():
            def _recurse(node, is_last, stack):
                stk = stack + [_WalkObj(node, False, is_last)]
                yield stk
                stk[-1].visited = True
                if isinstance(node, Group):
                    if show_options:
                        for o, l in _is_last_iter(node.options.values()):
                            yield from _recurse(o, l, stk)
                    for g, l in _is_last_iter(node.groups.values()):
                        yield from _recurse(g, l, stk)
            return _recurse(self, True, [])

        if colors:
            nG, nO, S, G, O, R = '\033[35m', '\033[33m', '\033[90m', '\033[95m', '\033[93m', '\033[0m'
        else:
            nG, nO, S, G, O, R = '', '', '', '', '', ''

        TREE = {
            (1, 1, 1, 0): f'   ',           # group,  visited,   last,  has groups after
            (1, 1, 1, 1): f'   ',           # group,  visited,   last,  no groups after
            (1, 0, 1, 0): f'   ',           # group,  unvisited, last,  has groups after
            (1, 1, 0, 1): f' {S}│{R} ',     # group,  visited,   inner, no groups after
            (1, 0, 0, 1): f' {S}├{G}─{R}',  # group,  unvisited, inner, no groups after
            (1, 0, 1, 1): f' {S}╰{G}─{R}',  # group,  unvisited, last,  no groups after
            (0, 0, 0, 1): f' {S}│{R} ',     # option, unvisited, last,  has groups after
            (0, 0, 1, 1): f' {S}├{O}╌{R}',  # option, unvisited, last,  no groups after
            (0, 0, 0, 0): f' {S}├{O}╌{R}',  # option, unvisited, inner, has groups after
            (0, 0, 1, 0): f' {S}╰{O}╌{R}',  # option, unvisited, last,  has groups after
        }

        for stack in _walk():
            (*_, item) = stack
            tree = ''.join(TREE.get(o.key, f'ERR{o}') for o in stack[1:])
            if item.is_group:
                keys = [f'/{k}' for k in (item.node.keys if item.node.keys else ('',))]
                name = f"{nG}{keys[-1]}{R}"
                if full_group_path:
                    name = f"{S}{''.join(keys[:-1])}{R}" + name
            else:
                name = f"{nO}{item.node.key}{R}"
                if full_option_path:
                    name = f"{S}{('/' + '/'.join(item.node.keys[:-1]))}:{R} " + name
            print(tree, name)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Schema                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def from_dict(cls, raw_group: dict, validate=True):
        if validate:
            raw_group = s.VerboseGroup.validate(raw_group)
        group = Group()
        for key, child in raw_group[s.KEY_CHILDREN].items():
            if child[s.KEY_TYPE] == s.TYPE_GROUP:
                group.add_subgroup(key, Group.from_dict(child, validate=False))
            elif child[s.KEY_TYPE] == s.TYPE_OPTION:
                group.add_option(key, Option.from_dict(child, validate=False))
            else:
                raise ValueError(f'Invalid type: {child[s.KEY_TYPE]}')
        return group

    def to_dict(self, validate=True):
        group = {
            s.KEY_TYPE: s.TYPE_GROUP,
            s.KEY_CHILDREN: {k: self[k].to_dict(validate=False) for k in self}
        }
        return s.VerboseGroup.validate(group) if validate else group


# ========================================================================= #
# Option                                                                    #
# ========================================================================= #


class Option(_ConfigObject):

    def __init__(
            self,
            data: dict = None,
            pkg: str = None,
            defaults: Dict[str, str] = None,
    ):
        super().__init__()
        self._data = data if data is not None else {}
        self._pkg = pkg if pkg is not None else s.DEFAULT_PKG
        self._defaults = defaults if defaults is not None else {}
        assert isinstance(self._pkg, (str, ConfigNode)), f'{s.KEY_PKG} is not a string or {ConfigNode.__name__}'
        assert isinstance(self._data, dict), f'{s.KEY_DATA} is not a dictionary'
        assert isinstance(self._defaults, dict), f'{s.KEY_MERGE} is not a dicitonary'
        # we dont validate in case things are nodes

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # getters                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @property
    def root(self) -> 'Group':
        assert self.has_parent
        g = super().root
        assert isinstance(g, Group)
        return g

    @property
    def group(self) -> Group:
        assert self.has_parent
        g = self.parent
        assert isinstance(g, Group)
        return g

    @property
    def group_keys(self) -> Tuple[str]:
        return self.group.keys

    @property
    def group_path(self):
        return self.group.group_path

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # getters - data                                                        #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    class _ReplaceStrings(RecursiveTransformer):
        def _transform_str(self, value):
            return SubNode(value)
        def __transform_default__(self, value):
            return value
        def _transform_dict_key(self, key):
            # do not allow interpolation on keys
            if isinstance(key, str):
                return key
            elif isinstance(key, ConfigNode):
                raise TypeError(f'keys in configs cannot be config nodes: {key}')
            return super()._transform_dict_key(key)

    def get_unresolved_defaults(self):
        return self._ReplaceStrings().transform(self._defaults)

    def get_unresolved_package(self):
        return self._ReplaceStrings().transform(self._pkg)

    def get_unresolved_data(self):
        return self._ReplaceStrings().transform(self._data)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Children - disabled for the option node                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    add_child = None
    get_child = None
    __setitem__ = None
    __getitem__ = None
    __iter__ = ().__iter__

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Strings                                                               #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'{self.__class__.__name__}({self._data})'

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # schema                                                                #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @classmethod
    def from_dict(cls, option, validate=True):
        if validate:
            option = s.VerboseOption.validate(option)
        return Option(
            pkg=option[s.KEY_PKG],
            defaults=option[s.KEY_MERGE],
            data=option[s.KEY_DATA],
        )

    def to_dict(self, validate=True):
        option = {
            s.KEY_TYPE: s.TYPE_OPTION,
            s.KEY_PKG: self._pkg,
            s.KEY_MERGE: self._defaults,
            s.KEY_DATA: self._data,
        }
        return s.VerboseOption.validate(option) if validate else option


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import keyword as _keyword
from schema import Schema as _Schema
from schema import Optional as _Optional
from schema import And as _And
from schema import Or as _Or
from schema import Use as _Use
from schema import Const as _Const
from schema import Regex as _Regex
from schema import SchemaError as _SchemaError


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _rename_fn(name, fn, debug=False):
    """
    Wraps a function in an instance of a class with __call__ defined.
    The new class has its __repr__ set to the passed name.
    - Useful for schema error messages
    """
    class wrap:
        def __call__(self, *args, **kwargs):
            return fn(*args, **kwargs)
        def __repr__(self):
            return f'{name}'
    if debug:
        return _debug_fn(wrap())
    return wrap()


def _debug_fn(fn):
    def wrap(*args, **kwargs):
        print(f'{repr(fn)} called with: args={args}, kwargs={kwargs}')
        result = fn(*args, **kwargs)
        print(f'- result: {result}')
        return result
    return wrap


# ========================================================================= #
# Keys                                                                      #
# ========================================================================= #


# keys - all
KEY_TYPE = '__type__'
# keys - options
KEY_PKG = '__package__'
KEY_MERGE = '__defaults__'
KEY_DATA = '__data__'
# keys - groups
KEY_CHILDREN = '__children__'

# marker keys - these are not reserved
MARKER_KEY_TARGET = '_target_'  # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED
MARKER_KEY_ARGS   = '_args_'    # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED
MARKER_KEY_KWARGS = '_kwargs_'  # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED
MARKER_KEY_VALUE  = '_value_'   # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED

# special types
TYPE_OPTION = 'option'
TYPE_GROUP  = 'group'
TYPE_NODE  = 'node'             # TODO: THIS IS NOT IMPLEMENTED, JUST RESERVED

# special packages
PKG_ROOT = '<root>'
PKG_GROUP = '<group>'

# special options
OPT_SELF = '<self>'

# default
DEFAULT_PKG = PKG_GROUP

# - - - - - - - - - - - - - #
# all values                #
# - - - - - - - - - - - - - #

# these keys are allowed as values anywhere and are not reserved
_MARKER_KEYS = {
    MARKER_KEY_TARGET,
    MARKER_KEY_ARGS,
    MARKER_KEY_KWARGS,
    MARKER_KEY_VALUE
}

_RESERVED_GROUP_KEYS = {
    KEY_TYPE, KEY_PKG, KEY_MERGE, KEY_DATA,
}

_RESERVED_OPTION_KEYS = {
    KEY_TYPE, KEY_CHILDREN,
}

# TODO: not yet implemented
_RESERVED_NODE_KEYS = {
    KEY_TYPE, KEY_DATA,
}

# these keys are not allowed in normal data
RESERVED_KEYS = {
    *_RESERVED_GROUP_KEYS,
    *_RESERVED_OPTION_KEYS,
    *_RESERVED_NODE_KEYS,
}

# ========================================================================= #
# Helper Types                                                              #
# ========================================================================= #


Value = _Schema(None)
Value._schema = _Or(
    int,
    float,
    str,
    tuple([Value]),
    list([Value]),
    dict({Value: Value}),
)

Identifier = _Schema(_And(
    str,
    _rename_fn('is_not_empty',       lambda key: bool(key)),
    _rename_fn('is_identifier',      lambda key: str.isidentifier(key)),
    _rename_fn('is_not_keyword',     lambda key: not _keyword.iskeyword(key)),
    _rename_fn('is_not_special_key', lambda key: key not in RESERVED_KEYS),
    _rename_fn('is_not_reserved',    lambda key: not (key.startswith('__') and key.endswith('__'))),  # we allow one underscore on each side for markers
), name='identifier')

IdentifierList = _Schema([Identifier], name='identifier_list')


def _split_path(path: str, sep: str):
    # check and remove prefix
    is_prefixed = False
    if path.startswith(sep):
        is_prefixed, path = True, path[len(sep):]
    # split path
    if path:
        split = str.split(path, sep)
    else:
        split = []
    # check identifiers
    return IdentifierList.validate(split), is_prefixed


def split_pkg_path(path: str):
    if path in {PKG_GROUP, PKG_ROOT}:
        raise RuntimeError('special package keys should be handled separately')
    keys, is_relative = _split_path(path, '.')
    return keys, is_relative


def split_group_path(path: str):
    keys, is_not_relative = _split_path(path, '/')
    return keys, not is_not_relative


PkgPath = _Schema(_Or(
    PKG_ROOT,
    PKG_GROUP,
    # sequentially convert and validate as a list of identifiers
    _And(str, _Const(split_pkg_path))
), name='package_path')

GroupPath = _Schema((
    # sequentially convert and validate as a list of identifiers
    _And(str, _Const(split_group_path))
), name='group_path')


# ========================================================================= #
# Sub Types                                                                 #
# ========================================================================= #


NameKey  = _Schema(Identifier, name='name_key')

PkgValue  = _Schema(_Or(PKG_ROOT, PKG_GROUP, PkgPath), name=KEY_PKG)
DataValue = _Schema({_Optional(Value): Value}, name=KEY_DATA)
MergersValue = _Schema({
    _Optional(OPT_SELF, default=None): None,
    _Optional(GroupPath): Identifier
}, name=KEY_MERGE)


# ========================================================================= #
# VERBOSE VERSIONS                                                          #
# - these versions are used internally                                      #
# ========================================================================= #


VerboseOption = _Schema({}, name='verbose_option')
VerboseOption.schema.update({
    _Optional(KEY_TYPE, default=TYPE_OPTION): TYPE_OPTION,
    _Optional(KEY_PKG, default=DEFAULT_PKG): PkgValue,
    _Optional(KEY_MERGE, default=dict):      MergersValue,
    _Optional(KEY_DATA, default=dict):       DataValue,
})

VerboseGroup = _Schema({}, name='verbose_group')
VerboseGroup.schema.update({
    _Optional(KEY_TYPE, default=TYPE_GROUP): TYPE_GROUP,
    _Optional(KEY_CHILDREN, default=list): {
        _Optional(NameKey): _Or(VerboseOption, VerboseGroup),
    },
})


# ========================================================================= #
# END                                                                       #
# ========================================================================= #




import pytest

from eunomia.backend import BackendObj, BackendDict, BackendYaml
from tests.test_backend_obj import _make_config_group
from tests.util import path_from_root


# ========================================================================= #
# Test Config Objects                                                       #
# ========================================================================= #


def test_simple_option():
    root = _make_config_group()

    # test group backend
    assert BackendObj(root).load_root_group().to_dict() == root.to_dict()
    with pytest.raises(TypeError, match='must be an instance of Group'):
        BackendObj(root.to_dict()).load_root_group().to_dict()

    # test dict backend
    assert BackendDict(root.to_dict()).load_root_group().to_dict() == root.to_dict()
    with pytest.raises(TypeError, match='must be a dict'):
        BackendDict(root).load_root_group().to_dict()

    # test yaml backend
    BackendYaml(path_from_root('examples/docs/quickstart/configs')).load_root_group()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



from glob import glob
import os
import ruamel.yaml as yaml

from eunomia.backend import Backend
from eunomia.config import Group, Option
from eunomia.config.nodes import IgnoreNode, RefNode, EvalNode, SubNode
from eunomia.config import scheme as s


# ========================================================================= #
# Config Backend                                                            #
# ========================================================================= #


class BackendYaml(Backend):

    def __init__(self, root_folder: str):
        # check the root folder
        if not isinstance(root_folder, str):
            raise TypeError(f'root_folder={repr(root_folder)} must be a string')
        if not os.path.isdir(root_folder):
            raise FileNotFoundError(f'root_folder={repr(root_folder)} is not a valid directory.\n\tAre you sure the path is correct and not relative?\n\tCurrent working directory is: {repr(os.getcwdb().decode())}')
        self._root_folder = root_folder        # loading modes

    @staticmethod
    def _norm_split_path(path, force_relative=True):
        path = os.path.normpath(path)
        parts, head = [], path
        while True:
            head, tail = os.path.split(head)
            if tail:
                parts.append(tail)
                continue
            elif head:
                parts.append(head)
            break
        parts = parts[::-1]
        recon_path = os.path.join(*parts) if parts else ''
        assert recon_path == path, f'reconstructed path {repr(recon_path)} is not equal to original {repr(path)}'
        if force_relative:
            assert not os.path.isabs(path)
        return parts

    def _get_sorted_paths(self):
        ext = '.yaml'
        # get all yaml files
        found_paths = glob(os.path.join(self._root_folder, f'**/*{ext}'), recursive=True)
        # remove root folder from names
        paths = []
        for path in found_paths:
            # get the relative path
            p = os.path.relpath(path, self._root_folder)
            # strip the extension
            p = os.path.splitext(p)[0]
            # split the path
            p = self._norm_split_path(p)
            # append the path and split path used to sort
            paths.append((path, p))
        # sort paths by length, then alphabetically
        return sorted(paths, key=lambda p: (len(p[1]), *p[1]))

    def _option_from_compact_dict(self, data):
        if s.KEY_DATA in data:
            return Option.from_dict(data)
        else:
            # pop all config values
            pkg = data.pop(s.KEY_PKG, None)
            merge = data.pop(s.KEY_MERGE, None)
            type = data.pop(s.KEY_TYPE, None)
            # check that we have nothing extra
            if any(k in s.RESERVED_KEYS for k in data.keys()):
                raise KeyError(f'A reserved key was found in a compact option dictionary: {list(k for k in data.keys() if k in s.RESERVED_KEYS)}')
            # no need to validate because of above
            return Option(pkg=pkg, defaults=merge, data=data)

    def _load_root_group(self) -> Group:
        root = Group()
        for path, (*subgroups, option_name) in self._get_sorted_paths():
            # add subgroups
            group = root.get_subgroups_recursive(subgroups, make_missing=True)
            # add option by loading file
            data = yaml_load_file(path)
            group.add_option(option_name, self._option_from_compact_dict(data))
        # done!
        return root


# ========================================================================= #
# Yaml Loader                                                               #
# ========================================================================= #


class EunomiaSafeLoader(yaml.SafeLoader):

    def __init__(self, *args, always_substitute_strings=True, **kwargs):
        super().__init__(*args, **kwargs)
        # custom config values
        self._always_substitute_strings = always_substitute_strings

    @classmethod
    def add_constructors(cls, tags: list, constructor):
        for tag in tags:
            cls.add_constructor(tag, constructor)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Custom Constructors                                                   #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def construct_custom_tuple(self, node: yaml.Node):
        if not isinstance(node, yaml.SequenceNode):
            raise yaml.YAMLError(f'tag {node.tag} for tuple(...) is not compatible with node: {node.__class__.__name__}. {repr(node.value)} must be a sequence/list')
        return tuple(self.construct_sequence(node))

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Node Constructors                                                     #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def construct_node_ignore(self, node: yaml.Node):
        if not isinstance(node, yaml.ScalarNode):
            raise TypeError(f'tag {node.tag} for {IgnoreNode.__name__} is not compatible with node: {node.__class__.__name__} which is not a scalar.')
        return IgnoreNode(self.construct_scalar(node))

    def construct_node_ref(self, node: yaml.Node):
        return RefNode(self.construct_yaml_str(node))

    def construct_node_eval(self, node: yaml.Node):
        return EvalNode(self.construct_yaml_str(node))

    def construct_node_sub(self, node: yaml.Node):
        return SubNode(self.construct_yaml_str(node))


EunomiaSafeLoader.add_constructors(['!tuple'], EunomiaSafeLoader.construct_custom_tuple)
EunomiaSafeLoader.add_constructors(['!str'], EunomiaSafeLoader.construct_node_ignore)
EunomiaSafeLoader.add_constructors(['!ref'], EunomiaSafeLoader.construct_node_ref)
EunomiaSafeLoader.add_constructors(['!expr'], EunomiaSafeLoader.construct_node_eval)
EunomiaSafeLoader.add_constructors(['!sub'], EunomiaSafeLoader.construct_node_sub)


# NOTE: unknown tags can be parsed
#       - https://github.com/kislyuk/yq/blob/2dd5cf39e18f2caf0c416e3da1c93a62bc801e0b/yq/loader.py#L57
# TODO: hydra adds the following implicit_resolvers & filters them
#       - float: https://github.com/omry/omegaconf/blob/ad593a527556583dfefa6d7dd34beec3f183a438/omegaconf/_utils.py#L106
#       - timestamp: https://github.com/omry/omegaconf/blob/ad593a527556583dfefa6d7dd34beec3f183a438/omegaconf/_utils.py#L120


# ========================================================================= #
# Yaml Util                                                                 #
# ========================================================================= #


def yaml_load(string):
    return yaml.load(string, EunomiaSafeLoader, version='1.2')


def yaml_load_file(path):
    with open(path, 'r') as f:
        return yaml_load(f.read())


# ========================================================================= #
# End                                                                       #
# ========================================================================= #



import pytest
from schema import SchemaError
from eunomia.config import scheme as s


# ========================================================================= #
# Test YAML & Custom Tags                                                   #
# ========================================================================= #


def _do_identifier_tests(schema, prefix=''):
    # check simple valid
    schema.validate(prefix + 'valid_id')
    with pytest.raises(SchemaError, match="is_not_special_key"):    schema.validate(prefix + '__type__')
    with pytest.raises(SchemaError, match="is_not_reserved"):       schema.validate(prefix + '__invalid_id__')
    schema.validate(prefix + '_okay_id_')
    with pytest.raises(SchemaError, match="is_identifier"):         schema.validate(prefix + '1234')
    with pytest.raises(SchemaError, match="is_not_keyword"):        schema.validate(prefix + 'with')
    with pytest.raises(SchemaError, match="should be instance of"): schema.validate(1234)


def test_identifier():
    _do_identifier_tests(s.Identifier)


def test_pkg_path():
    # check - absolute
    assert s.PkgPath.validate('') == ''
    assert s.PkgPath.validate('valid1') == 'valid1'
    assert s.PkgPath.validate('valid1.valid2') == 'valid1.valid2'
    with pytest.raises(SchemaError): s.PkgPath.validate('valid1.')
    # check - relative
    assert s.PkgPath.validate('.') == '.'
    assert s.PkgPath.validate('.valid1') == '.valid1'
    assert s.PkgPath.validate('.valid1.valid2') == '.valid1.valid2'
    with pytest.raises(SchemaError): s.PkgPath.validate('.valid1.')

    # check splits - absolute
    assert s.split_pkg_path('') == ([], False)
    assert s.split_pkg_path('valid1') == (['valid1'], False)
    assert s.split_pkg_path('valid1.valid2') == (['valid1', 'valid2'], False)
    with pytest.raises(SchemaError): s.split_pkg_path('valid1.')
    # check splits - relative
    assert s.split_pkg_path('.') == ([], True)
    assert s.split_pkg_path('.valid1') == (['valid1'], True)
    assert s.split_pkg_path('.valid1.valid2') == (['valid1', 'valid2'], True)
    with pytest.raises(SchemaError): s.split_pkg_path('.valid1.')

    # check identifiers, should be same as above
    _do_identifier_tests(s.PkgPath)
    _do_identifier_tests(s.PkgPath, prefix='valid1.')
    _do_identifier_tests(s.PkgPath, prefix='_marker_.')
    with pytest.raises(SchemaError, match='is_not_reserved'):
        _do_identifier_tests(s.PkgPath, prefix='__invalid1__.')

    # check group...
    with pytest.raises(SchemaError, match='is_identifier'):
        s.PkgPath.validate('valid1/valid2/valid3')

    # check special keys
    s.PkgPath.validate('<root>')
    s.PkgPath.validate('<group>')
    # split special keys
    with pytest.raises(RuntimeError): s.split_pkg_path('<root>')
    with pytest.raises(RuntimeError): s.split_pkg_path('<group>')


def test_group_path():
    # check - relative
    assert s.GroupPath.validate('') == ''
    assert s.GroupPath.validate('valid1') == 'valid1'
    assert s.GroupPath.validate('valid1/valid2') == 'valid1/valid2'
    with pytest.raises(SchemaError): s.GroupPath.validate('valid1/')
    # check - absolute
    assert s.GroupPath.validate('/') == '/'
    assert s.GroupPath.validate('/valid1') == '/valid1'
    assert s.GroupPath.validate('/valid1/valid2') == '/valid1/valid2'
    with pytest.raises(SchemaError): s.GroupPath.validate('/valid1/')

    # check splits - relative
    assert s.split_group_path('') == ([], True)
    assert s.split_group_path('valid1') == (['valid1'], True)
    assert s.split_group_path('valid1/valid2') == (['valid1', 'valid2'], True)
    with pytest.raises(SchemaError): s.split_group_path('valid1/')
    # check splits - absolute
    assert s.split_group_path('/') == ([], False)
    assert s.split_group_path('/valid1') == (['valid1'], False)
    assert s.split_group_path('/valid1/valid2') == (['valid1', 'valid2'], False)
    with pytest.raises(SchemaError): s.split_group_path('/valid1/')

    # check identifiers, should be same as above
    _do_identifier_tests(s.GroupPath)
    _do_identifier_tests(s.GroupPath, prefix='valid1/')
    _do_identifier_tests(s.GroupPath, prefix='_marker_/')
    with pytest.raises(SchemaError, match='is_not_reserved'):
        _do_identifier_tests(s.GroupPath, prefix='__invalid1__/')

    # check group...
    with pytest.raises(SchemaError, match='is_identifier'):
        s.GroupPath.validate('valid1.valid2.valid3')

    # check special keys
    with pytest.raises(SchemaError): s.GroupPath.validate('<root>')
    with pytest.raises(SchemaError): s.GroupPath.validate('<group>')
    # split special keys
    with pytest.raises(RuntimeError): s.split_pkg_path('<root>')
    with pytest.raises(RuntimeError): s.split_pkg_path('<group>')
